# coding: utf-8
from sqlalchemy import BigInteger, Boolean, CheckConstraint, Column, Date, DateTime, Float, ForeignKey, Integer, JSON, Numeric, String, Table, Text, UniqueConstraint, text, ForeignKeyConstraint
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declarative_base


Base = declarative_base()
metadata = Base.metadata


class Account(Base):
    __tablename__ = 'account'
    __table_args__ = (
        UniqueConstraint('provider', 'provider_principal'),
    )

    account_id = Column(String, primary_key=True)
    request_handler_id = Column(Numeric)
    version = Column(Numeric, nullable=False, server_default=text("1"))
    password_digest = Column(String)
    firstname = Column(String, nullable=False)
    lastname = Column(String, nullable=False)
    initials = Column(String)
    creationtimestamp = Column(String)
    modificationtimestamp = Column(String)
    preferredarc = Column(String)
    email = Column(String)
    executive = Column(String, nullable=False, server_default=text("'na'::character varying"))
    aliases = Column(String)
    inst_no = Column(Numeric, nullable=False, server_default=text("1"))
    provider = Column(String)
    provider_principal = Column(String)

    role = relationship('Role', secondary='account_role')


t_account_role = Table(
    'account_role', metadata,
    Column('role_no', ForeignKey('role.role_no'), nullable=False),
    Column('account_id', ForeignKey('account.account_id'), nullable=False)
)


class AlmaOusType(Base):
    __tablename__ = 'alma_ous_types'

    ous_type = Column(String, primary_key=True)
    description = Column(String)


t_alma_ous_view = Table(
    'alma_ous_view', metadata,
    Column('project_code', String),
    Column('sous', String),
    Column('gous', String),
    Column('mous', String),
    Column('schedblock_name', String)
)


class AlmaOus(Base):
    __tablename__ = 'alma_ouses'

    alma_ous_id = Column(String, primary_key=True)
    parent_ous_id = Column(ForeignKey('alma_ouses.alma_ous_id'))
    project_code = Column(ForeignKey('projects.project_code'))
    ous_type = Column(ForeignKey('alma_ous_types.ous_type'), nullable=False)
    schedblock_name = Column(String)

    alma_ous_type = relationship('AlmaOusType')
    parent_ous = relationship('AlmaOus', remote_side=[alma_ous_id], backref=backref('children_ouses', cascade="all, delete-orphan"))
    calibrations = relationship('Calibration')
    execution_blocks = relationship('ExecutionBlock')
    project = relationship('Project')


class AlmaSourceDatum(Base):
    __tablename__ = 'alma_source_data'

    alma_ous_id = Column(String, primary_key=True, nullable=False)
    field_name = Column(String, primary_key=True, nullable=False)
    ra = Column(Float(53))
    dec = Column(Float(53))
    integration_time = Column(Float(53))
    angular_resolution = Column(Float(53))
    largest_angular_scale = Column(Float(53))
    science_field = Column(Boolean, server_default=text("false"))

    spectral_windows = relationship('AlmaSpwDatum', secondary='alma_spw_sources', back_populates='sources')

    def __repr__(self):
        return "<AlmaSourceDatum#{alma_ous_id} {field_name} {ra} {dec} {integration_time} {angular_resolution} {largest_angular_scale}>" \
            .format(**self.__dict__)



class AlmaSpwDatum(Base):
    __tablename__ = 'alma_spw_data'

    alma_ous_id = Column(String, primary_key=True, nullable=False)
    spw_name = Column(String, primary_key=True, nullable=False)
    min_frequency = Column(Float(53))
    max_frequency = Column(Float(53))
    bandwidth = Column(Float(53))
    num_channels = Column(Integer)
    spectral_resolution = Column(Float(53))

    sources = relationship('AlmaSourceDatum', secondary='alma_spw_sources', back_populates='spectral_windows')

    def __repr__(self):
        return "<AlmaSpwDatum#{alma_ous_id} {spw_name} {min_frequency} {max_frequency} {bandwidth} {num_channels} {spectral_resolution}>" \
            .format(**self.__dict__)


t_alma_spw_sources = Table(
    'alma_spw_sources', metadata,
    Column('alma_ous_id', String, primary_key=True, nullable=False),
    Column('spw_name', String, primary_key=True, nullable=False),
    Column('field_name', String, primary_key=True, nullable=False),
    ForeignKeyConstraint(['alma_ous_id', 'field_name'], ['alma_source_data.alma_ous_id', 'alma_source_data.field_name']),
    ForeignKeyConstraint(['alma_ous_id', 'spw_name'], ['alma_spw_data.alma_ous_id', 'alma_spw_data.spw_name'])
)


t_alma_project_codes = Table(
    'alma_project_codes', metadata,
    Column('code', String),
    Column('uid', String)
)


class AlmaReingestionQueue(Base):
    __tablename__ = 'alma_reingestion_queue'

    asdm_uid = Column(String, primary_key=True)
    observation_finished = Column(DateTime(True), nullable=False)
    last_updated = Column(DateTime(True), nullable=False, server_default=text("now()"))
    state = Column(String, nullable=False, server_default=text("'WAITING'::character varying"))


# class AlmaSourceDatum(Base):
#     __tablename__ = 'alma_source_data'
#
#     alma_ous_id = Column(String, primary_key=True, nullable=False)
#     field_name = Column(String, primary_key=True, nullable=False)
#     ra = Column(Float(53))
#     dec = Column(Float(53))
#     integration_time = Column(Float(53))
#     angular_resolution = Column(Float(53))
#     largest_angular_scale = Column(Float(53))
#
#     alma_ouss = relationship('AlmaSpwDatum', secondary='alma_spw_sources')


# class AlmaSpwDatum(Base):
#     __tablename__ = 'alma_spw_data'
#
#     alma_ous_id = Column(String, primary_key=True, nullable=False)
#     spw_number = Column(Integer, primary_key=True, nullable=False)
#     min_frequency = Column(Float(53))
#     max_frequency = Column(Float(53))
#     bandwidth = Column(Float(53))
#     num_channels = Column(Integer)
#     spectral_resolution = Column(Float(53))


# t_alma_spw_sources = Table(
#     'alma_spw_sources', metadata,
#     Column('alma_ous_id', String, primary_key=True, nullable=False),
#     Column('spw_number', Integer, primary_key=True, nullable=False),
#     Column('field_name', String, primary_key=True, nullable=False),
#     ForeignKeyConstraint(['alma_ous_id', 'field_name'], ['alma_source_data.alma_ous_id', 'alma_source_data.field_name']),
#     ForeignKeyConstraint(['alma_ous_id', 'spw_number'], ['alma_spw_data.alma_ous_id', 'alma_spw_data.spw_number'])
# )


class AncillaryProduct(Base):
    __tablename__ = 'ancillary_products'

    ancillary_product_locator = Column(String, primary_key=True)
    ancillary_product_type = Column(String, nullable=False)
    filegroup_id = Column(ForeignKey('filegroups.filegroup_id'), nullable=False)
    product_group_id = Column(ForeignKey('product_groups.product_group_id'))
    science_product_locator = Column(ForeignKey('science_products.science_product_locator'))

    filegroup = relationship('Filegroup')
    product_group = relationship('ProductGroup')
    science_product = relationship('ScienceProduct', backref=backref('ancillary_products'))

    @property
    def locator(self):
        return self.ancillary_product_locator

    @property
    def subdirectory(self):
        return ''


class Archive(Base):
    __tablename__ = 'archive'

    id = Column(Numeric, primary_key=True)
    name = Column(String(10), nullable=False)


class ArchiveLog(Base):
    __tablename__ = 'archive_log'
    __table_args__ = (
        UniqueConstraint('request_id', 'dataset_name'),
    )

    log_id = Column(Numeric, primary_key=True, server_default=text("nextval('log_sequence'::regclass)"))
    ip = Column(String)
    host_name = Column(String)
    account_id = Column(Numeric)
    country = Column(String)
    local_host_name = Column(String)
    user_agent = Column(String)
    start_date = Column(DateTime, nullable=False, server_default=text("CURRENT_TIMESTAMP"))
    program = Column(String, nullable=False)
    version = Column(String)
    resource_name = Column(String)
    query = Column(String)
    size_lines = Column(Numeric)
    size_bytes_wire = Column(Numeric)
    size_bytes_uncompressed = Column(Numeric)
    duration = Column(Numeric)
    medium = Column(String)
    user_category = Column(Numeric)
    pi_request = Column(String(1))
    member_state = Column(String(1))
    eso_user = Column(String(1))
    error_code = Column(Numeric)
    error_string = Column(String)
    file_id = Column(String)
    archive_class = Column(String)
    service_type = Column(String)
    dataset_name = Column(String)
    request_id = Column(Numeric)
    archive_id = Column(ForeignKey('archive.id'))

    archive = relationship('Archive')


class Author(Base):
    __tablename__ = 'authors'

    author_id = Column(Integer, primary_key=True, server_default=text("nextval('authors_author_id_seq'::regclass)"))
    project_code = Column(ForeignKey('projects.project_code', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)
    username = Column(String, nullable=False)
    firstname = Column(String, nullable=False)
    lastname = Column(String, nullable=False)
    pst_person_id = Column(String)
    is_pi = Column(Boolean, nullable=False)

    project = relationship('Project')

    def __repr__(self):
        return "<Author#{author_id} {username} '{firstname} {lastname}'>" \
            .format(**self.__dict__)


class CalibrationStatusValue(Base):
    __tablename__ = 'calibration_status_values'

    status = Column(String, primary_key=True)
    description = Column(String, nullable=False)


class Calibration(Base):
    __tablename__ = 'calibrations'

    calibration_id = Column(Integer, primary_key=True, server_default=text("nextval('calibrations_calibration_id_seq'::regclass)"))
    alma_ous_id = Column(ForeignKey('alma_ouses.alma_ous_id'))
    project_code = Column(ForeignKey('projects.project_code'), nullable=False)
    execution_block_id = Column(ForeignKey('execution_blocks.execution_block_id'))
    filegroup_id = Column(ForeignKey('filegroups.filegroup_id'), nullable=False)
    science_product_locator = Column(ForeignKey('science_products.science_product_locator'), unique=True)

    alma_ous = relationship('AlmaOus')
    execution_block = relationship('ExecutionBlock')
    filegroup = relationship('Filegroup')
    project = relationship('Project')
    science_product = relationship('ScienceProduct', uselist=False)

    def __repr__(self):
        return "<Calibration#{calibration_id} {execution_block_id} '{alma_ous_id} {filegroup_id} {project_code}'>" \
            .format(**self.__dict__)

class VLASSCalibration(Calibration):
    __tablename__ = 'VLASS_calibrations'

    calibration_id = Column(ForeignKey('calibrations.calibration_id'), primary_key=True)


class Catalog(Base):
    __tablename__ = 'catalogs'

    catalog_id = Column(Integer, primary_key=True, server_default=text("nextval('catalogs_catalog_id_seq'::regclass)"))
    science_product_locator = Column(ForeignKey('science_products.science_product_locator'), nullable=False)

    science_product = relationship('ScienceProduct')


class VlassCatalog(Catalog):
    __tablename__ = 'vlass_catalogs'

    catalog_id = Column(ForeignKey('catalogs.catalog_id'), primary_key=True)
    tile = Column(String)
    type = Column(String)


class Configuration(Base):
    __tablename__ = 'configurations'

    configuration_id = Column(Integer, primary_key=True, server_default=text("nextval('configurations_configuration_id_seq'::regclass)"))
    configuration = Column(Integer, nullable=False)
    execution_block_id = Column(ForeignKey('execution_blocks.execution_block_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False, index=True)

    execution_block = relationship('ExecutionBlock')

    def __repr__(self):
        return "<Configuration#{configuration_id}>".format(**self.__dict__)



class DataDescription(Base):
    __tablename__ = 'data_descriptions'

    data_description_id = Column(Integer, primary_key=True, server_default=text("nextval('data_descriptions_data_description_id_seq'::regclass)"))
    bandwidth = Column(Float(53), nullable=False)
    frequency = Column(Float(53), nullable=False)
    polarization_id = Column(ForeignKey('polarizations.polarization_id'), nullable=False)
    configuration_id = Column(ForeignKey('configurations.configuration_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False, index=True)

    configuration = relationship('Configuration')
    polarization = relationship('Polarization')
    subscans = relationship('Subscan', secondary='subscan_data_descriptions')

    def __repr__(self):
        return "<DataDescription#{data_description_id} Freq={frequency} BW={bandwidth}>" \
            .format(**self.__dict__)

class Databasechangelog(Base):
    __tablename__ = 'databasechangelog'

    id = Column(String(63), primary_key=True, nullable=False)
    author = Column(String(63), primary_key=True, nullable=False)
    filename = Column(String(200), primary_key=True, nullable=False)
    dateexecuted = Column(DateTime(True), nullable=False)
    orderexecuted = Column(Integer, nullable=False)
    exectype = Column(String(10), nullable=False)
    md5sum = Column(String(35))
    description = Column(String(255))
    comments = Column(String(255))
    tag = Column(String(255))
    liquibase = Column(String(20))
    contexts = Column(String(255))
    labels = Column(String(255))
    deployment_id = Column(String(10))


class Databasechangeloglock(Base):
    __tablename__ = 'databasechangeloglock'

    id = Column(Integer, primary_key=True)
    locked = Column(Boolean, nullable=False)
    lockgranted = Column(DateTime(True))
    lockedby = Column(String(255))


class Event(Base):
    __tablename__ = 'events'

    id = Column(Integer, primary_key=True, server_default=text("nextval('events_id_seq'::regclass)"))
    application = Column(String, nullable=False)
    user = Column(String)
    request = Column(String)
    message = Column(String, nullable=False)
    log_data = Column(JSON)
    version = Column(Float(53), nullable=False)
    timestamp = Column(DateTime(True), nullable=False)

    def __repr__(self):
        return "<Event#{id} app={application} user={user} request={request} timestamp={timestamp}>" \
            .format(**self.__dict__)


t_execblock_start_stop = Table(
    'execblock_start_stop', metadata,
    Column('execution_block_id', Integer),
    Column('starttime', Float(53)),
    Column('endtime', Float(53))
)


class ExecutionBlock(Base):
    __tablename__ = 'execution_blocks'
    __table_args__ = (
        CheckConstraint("(((telescope)::text = 'EVLA'::text) AND (ngas_fileset_id IS NOT NULL)) OR ((telescope)::text <> 'EVLA'::text)"),
    )

    execution_block_id = Column(Integer, primary_key=True, server_default=text("nextval('execution_blocks_execution_block_id_seq'::regclass)"))
    ost_exec_block_id = Column(Integer)
    filegroup_id = Column(ForeignKey('filegroups.filegroup_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)
    calibration_level = Column(String, nullable=False)
    telescope = Column(String, nullable=False)
    configuration = Column(String)
    scheduling_block_id = Column(Integer)
    ngas_fileset_id = Column(String)
    project_code = Column(ForeignKey('projects.project_code', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)
    starttime = Column(Float(53))
    endtime = Column(Float(53))
    calibration_status = Column(ForeignKey('calibration_status_values.status'), nullable=False, server_default=text("'Unknown'::character varying"))
    scheduling_block_type = Column(String)
    band_code = Column(String)
    ingestion_complete = Column(Boolean, nullable=False, server_default=text("false"))
    segment = Column(String(2))
    alma_ous_id = Column(ForeignKey('alma_ouses.alma_ous_id'))
    science_product_locator = Column(ForeignKey('science_products.science_product_locator'), unique=True)

    alma_ous = relationship('AlmaOus')
    calibration_status_value = relationship('CalibrationStatusValue')
    filegroup = relationship('Filegroup')
    project = relationship('Project')
    science_product = relationship('ScienceProduct', uselist=False)
    scans = relationship('Scan', cascade="all, delete-orphan")
    calibrations = relationship('Calibration')

    def __repr__(self):
        return "<ExecutionBlock#{execution_block_id} project={project_code} start={starttime} end={endtime}>".format(**self.__dict__)


class RealfastExecutionBlock(ExecutionBlock):
    __tablename__ = 'realfast_execution_blocks'

    execution_block_id = Column(ForeignKey('execution_blocks.execution_block_id'), primary_key=True)
    transient_ra = Column(String)
    transient_ra_error = Column(Float(53))
    transient_dec = Column(String)
    transient_dec_error = Column(Float(53))
    transient_snr = Column(Float(53))
    transient_dm = Column(Float(53))
    transient_dm_error = Column(Float(53))
    preaverage_time = Column(Float(53))
    rfpipe_version = Column(String)
    prefs_id = Column(String)
    rf_qa_label = Column(String)
    rf_qa_zero_fraction = Column(Float(53))
    rf_qa_visibility_noise = Column(Float(53))
    rf_qa_image_noise = Column(Float(53))


class ExternalProductSystem(Base):
    __tablename__ = 'external_product_systems'

    name = Column(String, primary_key=True)

    def __repr__(self):
        return "<ExternalProductSystem#{name}>".format(**self.__dict__)


class VLASSExecutionBlock(ExecutionBlock):
    __tablename__ = 'VLASS_execution_blocks'

    execution_block_id = Column(ForeignKey('execution_blocks.execution_block_id'), primary_key=True)


class FailedIngestion(Base):
    __tablename__ = 'failed_ingestions'

    filename = Column(String(), primary_key=True)
    telescope = Column(String())
    last_attempted = Column(DateTime(True))
    category = Column(String())
    failure_reason = Column(String())
    resolution = Column(String())
    state = Column(String())


class Filegroup(Base):
    __tablename__ = 'filegroups'
    __table_args__ = (
        CheckConstraint('(CASE WHEN (project_code IS NOT NULL) THEN 1 ELSE 0 END + CASE WHEN (parent_filegroup_id IS NOT NULL) THEN 1 ELSE 0 END) = 1'),
    )

    filegroup_id = Column(Integer, primary_key=True, server_default=text("nextval('filegroups_filegroup_id_seq'::regclass)"))
    project_code = Column(ForeignKey('projects.project_code', ondelete='CASCADE', onupdate='CASCADE'))
    parent_filegroup_id = Column(ForeignKey('filegroups.filegroup_id', ondelete='CASCADE', onupdate='CASCADE'))
    datasize = Column(BigInteger)
    type = Column(String)

    parent_filegroup = relationship('Filegroup', remote_side=[filegroup_id], backref=backref('children_filegroups', cascade="all, delete-orphan"))
    project = relationship('Project')

    execution_blocks = relationship('ExecutionBlock', cascade="all, delete-orphan")
    files = relationship('File', cascade="all, delete-orphan")
    scans = relationship('Scan')

    @property
    def all_files(self):
        """
        :return: all the files under this filegroup, recursively.

        Beware performance issues here!
        """
        # this is a gross and probably expensive albeit beautiful way to write this query
        return self.files + [file for subgroup in self.children_filegroups for file in subgroup.all_files]

    def __repr__(self):
        return "<Filegroup#{filegroup_id} project={project_code}>".format(**self.__dict__)


class File(Base):
    __tablename__ = 'files'

    file_id = Column(Integer, primary_key=True, server_default=text("nextval('files_file_id_seq'::regclass)"))
    file_path = Column(String)
    ngas_id = Column(String)
    filegroup = Column(ForeignKey('filegroups.filegroup_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False, index=True)
    filename = Column(String, nullable=False)
    filesize = Column(BigInteger, nullable=False)
    format = Column(String, nullable=False)
    type = Column(String, nullable=False)
    checksum = Column(String)
    checksum_type = Column(String)
    ingestion_time = Column(DateTime(True))
    ngas_cluster = Column(String)
    ngas_location = Column(String)
    preview_storage_path = Column(String)

    filegroup1 = relationship('Filegroup')
    images = relationship('Image', cascade="all, delete-orphan")
    subscans = relationship('Subscan')

    def __repr__(self):
        return "<File#{file_id} {file_path}/{filename} ngas_id={ngas_id}>".format(**self.__dict__)

class ImageProduct(Base):
    __tablename__ = 'image_products'

    image_product_id = Column(Integer, primary_key=True, server_default=text("nextval('image_products_image_product_id_seq'::regclass)"))
    project_code = Column(ForeignKey('projects.project_code'), nullable=False)
    configurations = Column(String(255))
    collection_name = Column(String(255))
    calibration_level = Column(Integer, server_default=text("2"))
    product_file_id = Column(ForeignKey('files.file_id', ondelete='CASCADE', onupdate='CASCADE'), ForeignKey('files.file_id'), nullable=False)
    tags = Column(String(255))
    image_set_filegroup = Column(ForeignKey('filegroups.filegroup_id', ondelete='CASCADE', onupdate='CASCADE'))

    filegroup = relationship('Filegroup')
    product_file = relationship('File', primaryjoin='ImageProduct.product_file_id == File.file_id')
    product_file1 = relationship('File', primaryjoin='ImageProduct.product_file_id == File.file_id')
    project = relationship('Project')


class Image(Base):
    __tablename__ = 'images'

    image_id = Column(Integer, primary_key=True, server_default=text("nextval('images_image_id_seq'::regclass)"))
    file_id = Column(ForeignKey('files.file_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)
    target_name = Column(String, nullable=False)
    telescope = Column(String, nullable=False)
    thumbnail = Column(String)
    spatial_resolution = Column(Float(53), nullable=False)
    image_field_of_view = Column(Float(53), nullable=False)
    max_intensity = Column(Float(53), nullable=False)
    min_intensity = Column(Float(53), nullable=False)
    rms_noise = Column(Float(53), nullable=False)
    polarization_id = Column(ForeignKey('polarizations.polarization_id'), nullable=False)
    ra = Column(String(255), nullable=False)
    dec = Column(String(255), nullable=False)
    min_frequency = Column(Float(53), nullable=False)
    max_frequency = Column(Float(53), nullable=False)
    ra_element_count = Column(Integer, nullable=False)
    dec_element_count = Column(Integer, nullable=False)
    starttime = Column(Float(53))
    endtime = Column(Float(53))
    exposure_time = Column(Float(53))
    rest_frequency = Column(Float(53))
    image_units = Column(String(255))
    spatial_region = Column(Text)
    beam_axis_ratio = Column(Float(53))
    band_code = Column(String)
    ra_pixel_size = Column(Float(53), nullable=False)
    dec_pixel_size = Column(Float(53), nullable=False)
    tags = Column(String)
    science_product_locator = Column(ForeignKey('science_products.science_product_locator'), unique=True)
    collection_name = Column(String)
    configurations = Column(String)
    calibration_level = Column(Integer)

    file = relationship('File')
    polarization = relationship('Polarization')
    science_product = relationship('ScienceProduct', uselist=False)

    def __repr__(self):
        return "<Image#{image_id} target={target_name} telescope={telescope}>".format(**self.__dict__)

class VlassImage(Image):
    __tablename__ = 'vlass_images'

    image_id = Column(ForeignKey('images.image_id'), primary_key=True)
    epoch = Column(ForeignKey('vlass_epochs.epoch'))
    tile = Column(String)
    type = Column(ForeignKey('vlass_image_types.type'))

    vlass_epoch = relationship('VlassEpoch')
    vlass_image_type = relationship('VlassImageType')


class Intent(Base):
    __tablename__ = 'intents'

    intent_name = Column(String, primary_key=True)

    subscans = relationship('Subscan', secondary='subscan_intents')

    def __repr__(self):
        return "<Intent {0}>".format(self.intent_name)


t_logs = Table(
    'logs', metadata,
    Column('filename', String),
    Column('class', String),
    Column('method', String),
    Column('line', Integer),
    Column('arguments', String),
    Column('timestamp', DateTime(True)),
    Column('formatted_message', String),
    Column('logger_name', String),
    Column('level', String),
    Column('thread_name', String),
    Column('reference_mask', Integer),
    Column('properties', JSON),
    Column('stacktrace', Text),
    Column('request_id', Integer)
)


class Polarization(Base):
    __tablename__ = 'polarizations'

    polarization_id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)

    def __repr__(self):
        return "<Polarization {0}>".format(self.name)


class ProductGroup(Base):
    __tablename__ = 'product_groups'

    product_group_id = Column(Integer, primary_key=True, server_default=text("nextval('product_groups_product_group_id_seq'::regclass)"))
    product_group_type = Column(String, nullable=False)
    alma_ous_id = Column(ForeignKey('alma_ouses.alma_ous_id'))
    program_id = Column(ForeignKey('programs.program_id'))
    parent_product_group_id = Column(ForeignKey('product_groups.product_group_id'))

    ALMA_ous = relationship('AlmaOus')
    parent_product_group = relationship('ProductGroup', remote_side=[product_group_id])
    program = relationship('Program')
    science_products = relationship('ScienceProduct', secondary='science_products_product_groups')


class Program(Base):
    __tablename__ = 'programs'

    program_id = Column(String, primary_key=True)
    project_code = Column(ForeignKey('projects.project_code'), nullable=False)
    program_type = Column(String, nullable=False)
    parent_program_id = Column(ForeignKey('programs.program_id'))

    parent_program = relationship('Program', remote_side=[program_id])
    project = relationship('Project')


class Project(Base):
    __tablename__ = 'projects'

    project_code = Column(String, primary_key=True)
    legacy_id = Column(String)
    total_observation_time = Column(Float(53))
    opt_project_id = Column(Integer)
    title = Column(String)
    abstract = Column(Text)
    proprietary_expiration = Column(DateTime)
    last_addition = Column(Date)
    starttime = Column(Float(53))
    endtime = Column(Float(53))
    proprietary_duration = Column(Float(53))

    science_products = relationship('ScienceProduct', secondary='science_products_projects')
    authors = relationship('Author', cascade="all,delete, delete-orphan",
                           backref='parent')
    execution_blocks = relationship('ExecutionBlock')
    file_groups = relationship('Filegroup')
    alma_ouses = relationship('AlmaOus', backref='projects')

    def __repr__(self):
        return '<Project#{project_code} "{title}" start={starttime} end={endtime}>'.format(**self.__dict__)


class Receiver(Base):
    __tablename__ = 'receivers'

    receiver_id = Column(Integer, primary_key=True, server_default=text("nextval('receivers_receiver_id_seq'::regclass)"))
    description = Column(String, nullable=False)

    def __repr__(self):
        return '<Receiver {0}>'.format(self.receiver_id)


class ReingestionState(Base):
    __tablename__ = 'reingestion_states'

    state = Column(String, primary_key=True)


class RhCommand(Base):
    __tablename__ = 'rh_commands'

    command_id = Column(Numeric, primary_key=True)
    name = Column(String, nullable=False, unique=True)
    arguments = Column(String)
    description = Column(String)
    host = Column(String)
    script = Column(String, nullable=False)
    staging_area = Column(String)
    execution_level = Column(String, nullable=False)


class RhDataEntity(Base):
    __tablename__ = 'rh_data_entities'
    __table_args__ = (
        UniqueConstraint('data_entity_name', 'request_id'),
    )

    data_entity_id = Column(Numeric, primary_key=True)
    data_entity_name = Column(String, nullable=False)
    request_id = Column(ForeignKey('rh_requests.request_id'), nullable=False, index=True)
    state = Column(String, nullable=False)
    meta_data_uri = Column(String)

    request = relationship('RhRequest')


class RhFile(Base):
    __tablename__ = 'rh_files'
    __table_args__ = (
        CheckConstraint('((request_id IS NOT NULL) AND (data_entity_id IS NULL)) OR ((request_id IS NULL) AND (data_entity_id IS NOT NULL))'),
    )

    file_id = Column(Numeric, primary_key=True)
    request_id = Column(ForeignKey('rh_requests.request_id'), index=True)
    data_entity_id = Column(ForeignKey('rh_data_entities.data_entity_id'), index=True)
    file_key = Column(String, nullable=False, index=True)
    namespace = Column(String, nullable=False)
    file_name = Column(String, nullable=False)
    file_type = Column(String)
    file_uri = Column(String, nullable=False)
    file_size = Column(Numeric, nullable=False)
    rank = Column(Numeric)
    permission = Column(String)
    medium_id = Column(Numeric)

    data_entity = relationship('RhDataEntity')
    request = relationship('RhRequest')


class RhOperatorRequest(Base):
    __tablename__ = 'rh_operator_requests'

    request_id = Column(Numeric, primary_key=True)
    creation_date = Column(DateTime, nullable=False)
    priority = Column(String, nullable=False)
    resolved_flag = Column(String(1), nullable=False)
    action_id = Column(Numeric)
    request_description = Column(String)


class RhProcess(Base):
    __tablename__ = 'rh_processes'
    __table_args__ = (
        CheckConstraint('((request_id IS NOT NULL) AND (data_entity_id IS NULL)) OR ((request_id IS NULL) AND (data_entity_id IS NOT NULL))'),
    )

    process_id = Column(Numeric, primary_key=True)
    state = Column(String, nullable=False)
    command_id = Column(ForeignKey('rh_commands.command_id'), nullable=False)
    job_id = Column(String)
    delivery_method = Column(String)
    arguments = Column(String)
    file_id = Column(ForeignKey('rh_files.file_id'))
    request_id = Column(ForeignKey('rh_requests.request_id'))
    data_entity_id = Column(ForeignKey('rh_data_entities.data_entity_id'))
    process_type = Column(String, nullable=False)

    command = relationship('RhCommand')
    data_entity = relationship('RhDataEntity')
    file = relationship('RhFile')
    request = relationship('RhRequest')


class RhRequestParameter(Base):
    __tablename__ = 'rh_request_parameters'

    request_id = Column(ForeignKey('rh_requests.request_id'), primary_key=True, nullable=False)
    pname = Column(String, primary_key=True, nullable=False)
    pvalue = Column(String)

    request = relationship('RhRequest')


class RhRequest(Base):
    __tablename__ = 'rh_requests'

    request_id = Column(Numeric, primary_key=True)
    account_id = Column(Numeric, nullable=False)
    request_state = Column(String, nullable=False)
    creation_date = Column(DateTime, nullable=False)
    request_description = Column(String)
    delivery_method_type = Column(String)
    delivery_media_type = Column(String)
    delivery_media_address = Column(String)
    completion_date = Column(DateTime)
    priority_code = Column(Numeric, nullable=False)
    ds_file_size = Column(Numeric)
    aux_file_size = Column(Numeric)
    total_files_number = Column(Numeric)
    total_processes_number = Column(Numeric)
    notification_email = Column(String)
    archive_id = Column(ForeignKey('archive.id'))

    archive = relationship('Archive')


t_rh_requests_mediums = Table(
    'rh_requests_mediums', metadata,
    Column('request_id', ForeignKey('rh_requests.request_id'), primary_key=True, nullable=False),
    Column('medium_id', ForeignKey('rh_site_delivery_medium.medium_id'), primary_key=True, nullable=False)
)


class RhSiteDeliveryMedium(Base):
    __tablename__ = 'rh_site_delivery_medium'

    medium_id = Column(Numeric, primary_key=True)
    active = Column(String(1), nullable=False)
    media_type = Column(String, nullable=False)
    host_name = Column(String, nullable=False)
    device = Column(String, nullable=False)
    capacity = Column(Float, nullable=False)
    request_id = Column(Numeric)
    state = Column(String, nullable=False)
    last_mod = Column(DateTime, server_default=text("'2018-04-18 15:21:15.783814'::timestamp without time zone"))
    status_message = Column(String, nullable=False)

    requests = relationship('RhRequest', secondary='rh_requests_mediums')


class Role(Base):
    __tablename__ = 'role'
    __table_args__ = (
        UniqueConstraint('application', 'name'),
    )

    role_no = Column(Numeric, primary_key=True)
    version = Column(Numeric, nullable=False)
    application = Column(String, nullable=False)
    name = Column(String, nullable=False)
    parent_role = Column(ForeignKey('role.role_no'))

    parent = relationship('Role', remote_side=[role_no])


t_rsl_eb_cal_map = Table(
    'rsl_eb_cal_map', metadata,
    Column('eb_locator', String),
    Column('cal_locator', String)
)


class Scan(Base):
    __tablename__ = 'scans'

    scan_id = Column(Integer, primary_key=True, server_default=text("nextval('scans_scan_id_seq'::regclass)"))
    ost_scan_id = Column(Integer)
    execution_block_id = Column(ForeignKey('execution_blocks.execution_block_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)
    filegroup_id = Column(ForeignKey('filegroups.filegroup_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)
    max_bandwidth = Column(Float(53), nullable=False)
    min_bandwidth = Column(Float(53), nullable=False)
    polarization_code = Column(Integer, nullable=False)
    max_frequency = Column(Float(53), nullable=False)
    min_frequency = Column(Float(53), nullable=False)
    filename = Column(String)

    execution_block = relationship('ExecutionBlock')
    filegroup = relationship('Filegroup')
    subscans = relationship('Subscan', cascade="all, delete-orphan")

    def __repr__(self):
        return '<Scan#{0}>'.format(self.scan_id)


class ScienceProductGroupType(Base):
    __tablename__ = 'science_product_group_types'

    science_product_group_type = Column(String, primary_key=True)
    description = Column(String)


class ScienceProductType(Base):
    __tablename__ = 'science_product_types'

    science_product_type = Column(String, primary_key=True)
    description = Column(String)


class ScienceProduct(Base):
    __tablename__ = 'science_products'

    science_product_locator = Column(String, primary_key=True)
    science_product_type = Column(ForeignKey('science_product_types.science_product_type'), nullable=False)
    metadata_ingestion_date = Column(DateTime, nullable=False)
    metadata_ingestion_version = Column(String, nullable=False)
    filegroup_id = Column(ForeignKey('filegroups.filegroup_id'), nullable=False)
    external_name = Column(String)
    external_system = Column(ForeignKey('external_product_systems.name'))

    external_product_system = relationship('ExternalProductSystem')
    filegroup = relationship('Filegroup')
    science_product_type1 = relationship('ScienceProductType')

    execution_block = relationship('ExecutionBlock', uselist=False, back_populates='science_product')
    project = relationship('Project', uselist=False, secondary='science_products_projects')

    @property
    def locator(self):
        return self.science_product_locator

    @property
    def subdirectory(self):
        if self.science_product_type == 'Execution Block':
            return self.execution_block.ngas_fileset_id
        else:
            return self.external_name


    def __repr__(self):
        return f"<ScienceProduct#{self.science_product_locator} type={self.science_product_type}>"

class ScienceProductsProductGroup(Base):
    __tablename__ = 'science_products_product_groups'

    science_product_locator = Column(ForeignKey('science_products.science_product_locator'), primary_key=True, nullable=False)
    product_group_id = Column(ForeignKey('product_groups.product_group_id'), primary_key=True, nullable=False)

    product_group = relationship('ProductGroup')
    science_product = relationship('ScienceProduct')


t_science_products_projects = Table(
    'science_products_projects', metadata,
    Column('science_product_locator', ForeignKey('science_products.science_product_locator'), primary_key=True, nullable=False),
    Column('project_code', ForeignKey('projects.project_code'), primary_key=True, nullable=False)
)


t_source_3c48 = Table(
    'source_3c48', metadata,
    Column('project_code', String),
    Column('execution_block_id', Integer),
    Column('ost_scan_id', Integer),
    Column('starttime', DateTime(True)),
    Column('ngas_id', String),
    Column('source', Text),
    Column('receiver_band', Text)
)


t_subscan_data_descriptions = Table(
    'subscan_data_descriptions', metadata,
    Column('subscan_id', ForeignKey('subscans.subscan_id'), primary_key=True, nullable=False),
    Column('data_description_id', ForeignKey('data_descriptions.data_description_id'), primary_key=True, nullable=False, index=True)
)


t_subscan_intents = Table(
    'subscan_intents', metadata,
    Column('intent_name', ForeignKey('intents.intent_name'), primary_key=True, nullable=False),
    Column('subscan_id', ForeignKey('subscans.subscan_id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True, nullable=False)
)


class Subscan(Base):
    __tablename__ = 'subscans'

    subscan_id = Column(Integer, primary_key=True, server_default=text("nextval('subscans_subscan_id_seq'::regclass)"))
    scan_id = Column(ForeignKey('scans.scan_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False)
    file_id = Column(ForeignKey('files.file_id', ondelete='CASCADE', onupdate='CASCADE'), index=True)
    ost_subscan_id = Column(Integer)
    obstype = Column(String, nullable=False)
    starttime = Column(Float(53), nullable=False)
    endtime = Column(Float(53), nullable=False)
    sourcename = Column(String, nullable=False)
    sourcetype = Column(String, nullable=False)
    ra = Column(Float(53), nullable=False)
    dec = Column(Float(53), nullable=False)
    exposure_time = Column(Float(53), nullable=False)
    integration_time = Column(Float(53), nullable=False)
    receiver_id = Column(ForeignKey('receivers.receiver_id'), nullable=False)
    backend = Column(String, nullable=False)
    intent = Column(String)
    configuration_id = Column(ForeignKey('configurations.configuration_id', ondelete='CASCADE', onupdate='CASCADE'), nullable=False, index=True)

    configuration = relationship('Configuration')
    file = relationship('File')
    receiver = relationship('Receiver')
    scan = relationship('Scan')
    data_descriptions = relationship('DataDescription', secondary='subscan_data_descriptions')
    intents = relationship('Intent', secondary='subscan_intents')

    def __repr__(self):
        return '<Subscan#{subscan_id} {obstype} start={starttime} end={endtime} ra={ra} dec={dec} backend={backend} intent={intent}>'.format(**self.__dict__)


t_total_file_sizes = Table(
    'total_file_sizes', metadata,
    Column('filegroup_id', Integer),
    Column('parent_filegroup_id', Integer),
    Column('filesize', Numeric)
)


class VlassCalibrationType(Base):
    __tablename__ = 'vlass_calibration_types'

    type = Column(String, primary_key=True)
    description = Column(String)


t_vlass_calibrations = Table(
    'vlass_calibrations', metadata,
    Column('calibration_id', ForeignKey('calibrations.calibration_id'), primary_key=True),
    Column('type', ForeignKey('vlass_calibration_types.type'))
)


class VlassEpoch(Base):
    __tablename__ = 'vlass_epochs'

    epoch = Column(String, primary_key=True)
    description = Column(String)

    execution_blocks = relationship('ExecutionBlock', secondary='vlass_execution_blocks')


class VlassExecutionBlockTile(Base):
    __tablename__ = 'vlass_execution_block_tiles'

    execution_block_id = Column(ForeignKey('vlass_execution_blocks.execution_block_id'), primary_key=True, nullable=False)
    tile = Column(String, primary_key=True, nullable=False)

    execution_block = relationship('VlassExecutionBlock')


class VlassExecutionBlock(Base):
    __tablename__ = 'vlass_execution_blocks'

    execution_block_id = Column(ForeignKey('execution_blocks.execution_block_id'), primary_key=True)
    epoch = Column(ForeignKey('vlass_epochs.epoch'))

    execution_block = relationship('ExecutionBlock')
    vlass_epoch = relationship('VlassEpoch')


class VlassImageType(Base):
    __tablename__ = 'vlass_image_types'

    type = Column(String, primary_key=True)
    description = Column(String)


class VlbaReingestionQueue(Base):
    __tablename__ = 'vlba_reingestion_queue'

    filename = Column(String, primary_key=True)
    observation_finished = Column(DateTime(True), nullable=False)
    last_updated = Column(DateTime(True), nullable=False, server_default=text("now()"))
    state = Column(ForeignKey('reingestion_states.state'), nullable=False, server_default=text("'WAITING'::character varying"))

    reingestion_state = relationship('ReingestionState')


class ProblemSeverityList(Base):
    __tablename__ = 'problem_severity_list'

    severity = Column(Integer, primary_key=True)
    description = Column(String)


class ProblemType(Base):
    __tablename__ = 'problem_types'

    problem_type_id = Column(Integer, primary_key=True, server_default=text("nextval('problem_types_problem_type_id_seq'::regclass)"))
    description = Column(String)


class ScienceProductComment(Base):
    __tablename__ = 'science_product_comments'

    comment_id = Column(Integer, primary_key=True, server_default=text("nextval('science_product_comments_comment_id_seq'::regclass)"))
    problem_type_id = Column(ForeignKey('problem_types.problem_type_id'), nullable=False)
    science_product_locator = Column(ForeignKey('science_products.science_product_locator'), nullable=False)
    severity = Column(ForeignKey('problem_severity_list.severity'), nullable=False)
    comment = Column(String)

    problem_type = relationship('ProblemType')
    science_product = relationship('ScienceProduct')
    problem_severity_list = relationship('ProblemSeverityList')

